#include "mount.h"
#include <stdio.h>
#include <stdlib.h>
#include <rpc/pmap_clnt.h>
#include <string.h>
#include <memory.h>
#include <sys/socket.h>
#include <netinet/in.h>

#ifndef SIG_PF
#define SIG_PF void(*)(int)
#endif

static void
mountprog_1(struct svc_req *rqstp, register SVCXPRT *transp)
{
printf("AUTH flavor %d base %p length %d\n", rqstp->rq_cred.oa_flavor,
rqstp->rq_cred.oa_base, rqstp->rq_cred.oa_length);
{
int i; unsigned char *s = rqstp->rq_cred.oa_base;
for (i=0; i < rqstp->rq_cred.oa_length; i++) printf(" %2.2x", s[i]);
printf("\n");
}
#if 0
if (rqstp->rq_cred.oa_length == 0) {
  printf("auth size 0, return error\n");
  svcerr_auth(transp, AUTH_TOOWEAK);
  return;
}
#endif
	union {
		dirpath mountproc_mnt_1_arg;
		dirpath mountproc_umnt_1_arg;
	} argument;
	char *result;
	xdrproc_t _xdr_argument, _xdr_result;
	char *(*local)(char *, struct svc_req *);

	switch (rqstp->rq_proc) {
	case MOUNTPROC_NULL:
		_xdr_argument = (xdrproc_t) xdr_void;
		_xdr_result = (xdrproc_t) xdr_void;
		local = (char *(*)(char *, struct svc_req *)) mountproc_null_1_svc;
		break;

	case MOUNTPROC_MNT:
		_xdr_argument = (xdrproc_t) xdr_dirpath;
		_xdr_result = (xdrproc_t) xdr_fhstatus;
		local = (char *(*)(char *, struct svc_req *)) mountproc_mnt_1_svc;
		break;

	case MOUNTPROC_DUMP:
		_xdr_argument = (xdrproc_t) xdr_void;
		_xdr_result = (xdrproc_t) xdr_mountlist;
		local = (char *(*)(char *, struct svc_req *)) mountproc_dump_1_svc;
		break;

	case MOUNTPROC_UMNT:
		_xdr_argument = (xdrproc_t) xdr_dirpath;
		_xdr_result = (xdrproc_t) xdr_void;
		local = (char *(*)(char *, struct svc_req *)) mountproc_umnt_1_svc;
		break;

	case MOUNTPROC_UMNTALL:
		_xdr_argument = (xdrproc_t) xdr_void;
		_xdr_result = (xdrproc_t) xdr_void;
		local = (char *(*)(char *, struct svc_req *)) mountproc_umntall_1_svc;
		break;

	case MOUNTPROC_EXPORT:
		_xdr_argument = (xdrproc_t) xdr_void;
		_xdr_result = (xdrproc_t) xdr_exportlist;
		local = (char *(*)(char *, struct svc_req *)) mountproc_export_1_svc;
		break;

	default:
		svcerr_noproc (transp);
		return;
	}
	memset ((char *)&argument, 0, sizeof (argument));
	if (!svc_getargs (transp, (xdrproc_t) _xdr_argument, (caddr_t) &argument)) {
		svcerr_decode (transp);
		return;
	}
	result = (*local)((char *)&argument, rqstp);
	if (result != NULL && !svc_sendreply(transp, (xdrproc_t) _xdr_result, result)) {
		svcerr_systemerr (transp);
	}
	if (!svc_freeargs (transp, (xdrproc_t) _xdr_argument, (caddr_t) &argument)) {
		fprintf (stderr, "%s", "unable to free arguments");
		exit (1);
	}
	return;
}

char *hostname;
char *dir_to_serve;

int
main (int argc, char **argv)
{
	register SVCXPRT *transp;

	pmap_unset (MOUNTPROG, MOUNTVERS);

  int sock;
  struct sockaddr_in addr;

  if (argc != 3) { printf("gimme <hostname> <dir to serve>\n"); return 1; }
  hostname = argv[1];
  dir_to_serve = argv[2];

  sock = socket(AF_INET, SOCK_DGRAM, 0);
  addr.sin_family = AF_INET;
  addr.sin_port = htons(2050);
  addr.sin_addr.s_addr = INADDR_ANY;
  if (sock == -1) { perror("socket"); exit(1); }
  if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1)
    { perror("bind"); exit(1); }

	transp = svcudp_create(sock); //RPC_ANYSOCK);
	if (transp == NULL) {
		fprintf (stderr, "%s", "cannot create udp service.");
		exit(1);
	}
	if (!svc_register(transp, MOUNTPROG, MOUNTVERS, mountprog_1, IPPROTO_UDP)) {
		fprintf (stderr, "%s", "unable to register (MOUNTPROG, MOUNTVERS, udp).");
		exit(1);
	}

	svc_run ();
	fprintf (stderr, "%s", "svc_run returned");
	exit (1);
	/* NOTREACHED */
}
