#include "files.h"
#include <stdlib.h>
#include <string.h>

static char *buf;
static unsigned long buf_size;
static unsigned long buf_maxsize;

static cookie_cache *cc;
static unsigned long cc_size, cc_maxsize;

unsigned long new_file(char *file)
{
  unsigned long ret;
  unsigned long len = strlen(file);
  /* we put in 'buf' the filename followed by its cookie_cache index */
  while (buf_size + len+1 + sizeof(long) > buf_maxsize) {
    buf_maxsize += 1024 * 1024 * 128;
    buf = realloc(buf, buf_maxsize); if (buf == NULL) abort();
  }
  memcpy(buf + buf_size + sizeof(long), file, len+1);
  *(long *)(buf + buf_size) = cc_size; /* not portable AT ALL */
  ret = buf_size;
  buf_size += len+1+sizeof(long);

  /* create a new empty cookie_cache */
  if (cc_size == cc_maxsize) {
    cc_maxsize += 1024;
    cc = realloc(cc, cc_maxsize * sizeof(cookie_cache));
    if (cc == NULL) abort();
  }
  bzero(&cc[cc_size], sizeof(cookie_cache));
  cc_size++;

  return ret;
}

void init_files(void)
{
  if (buf) free(buf);
  buf = NULL;
  buf_size = buf_maxsize = 0;
  new_file(".");
}

char *ul2name(unsigned long x)
{
  if (x + sizeof(long) >= buf_size) return NULL;
  return buf + x + sizeof(long);
}

unsigned long fhandle2ul(char *fhandle)
{
  /* non-portable, if you have an error, this might be from here */
  return *(unsigned long *)fhandle;
}

void ul2fhandle(unsigned long ul, char *fhandle)
{
  bzero(fhandle, 32);
  /* non-portable */
  *(unsigned long *)fhandle = ul;
}

char *fhandle2name(char *fhandle)
{
  return ul2name(fhandle2ul(fhandle));
}

cookie_cache *fhandle2cookies(char *fhandle)
{
  unsigned long h = fhandle2ul(fhandle);
  /* not portable AT ALL */
  unsigned long index = *(unsigned long *)(buf + h);
  return &cc[index];
}

long get_cookie(cookie_cache *c, int i)
{
  return c->d[i-1];
}

void clear_cookie_cache(cookie_cache *c)
{
  c->dsize = 0;
}

int add_cookie(cookie_cache *c, long off)
{
  if (c->dsize == c->dmaxsize) {
    c->dmaxsize += 128;
    c->d = realloc(c->d, c->dmaxsize * sizeof(long));
    if (c->d == NULL) abort();
  }
  c->d[c->dsize] = off;
  c->dsize++;
  return c->dsize;
}
