#include "nfs.h"
#include "files.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <utime.h>
#include <alloca.h>
#include <fcntl.h>
#include <sys/statvfs.h>
#include <dirent.h>

nfs_stat errno2status(int err)
{
printf("ERRNO %d %s\n", errno, strerror(errno));
fflush(stdout);
  enum nfs_stat ret;
  switch (err) {
  case EACCES: ret = NFSERR_ACCES; break;
  case ENAMETOOLONG: ret = NFSERR_NAMETOOLONG; break;
  case EFBIG: ret = NFSERR_FBIG; break;
  case ENOENT: ret = NFSERR_NOENT; break;
  case ENOTDIR: ret = NFSERR_NOTDIR; break;
  case EPERM: ret = NFSERR_PERM; break;
  case EROFS: ret = NFSERR_ROFS; break;
  case EISDIR: ret = NFSERR_ISDIR; break;
  case EDQUOT: ret = NFSERR_DQUOT; break;
  case ENXIO: ret = NFSERR_NXIO; break;
  case EEXIST: ret = NFSERR_EXIST; break;
  case ENODEV: ret = NFSERR_NODEV; break;
  case ENOSPC: ret = NFSERR_NOSPC; break;
  case ENOTEMPTY: ret = NFSERR_NOTEMPTY; break;
  case EIO:
  default: ret = NFSERR_IO; break;

  case EINTR: printf("we should not come here\n"); abort();
  }
  return ret;
}

int set_attributes(char *filename, sattr *attributes)
{
printf("set_attributes %s\n", filename);
printf("mode 0x%x uid %d gid %d size %d\n",
  attributes->mode, attributes->uid, attributes->gid,
  attributes->size);
printf("atime %d %d mtime %d %d\n",
  attributes->atime.seconds, attributes->atime.useconds,
  attributes->mtime.seconds, attributes->mtime.useconds);

printf("LINK? %d\n", S_ISLNK(attributes->mode));
  if (attributes->mode != (u_int)-1 && !S_ISLNK(attributes->mode))
    if (chmod(filename, attributes->mode) == -1) return -1;

printf("chmod ok\n");

  if (attributes->uid != (u_int)-1 || attributes->gid != (u_int)-1)
    if (lchown(filename, attributes->uid, attributes->gid) == -1)
      return -1;

printf("chown ok\n");

  if (attributes->size != (u_int)-1 && S_ISREG(attributes->mode))
    if (truncate(filename, attributes->size) == -1) {
printf("truncate error %s\n", strerror(errno));
return -1;
}

printf("truncate ok\n");

  /* not sure of the tests here */
#if 0
  if (attributes->atime.seconds != (u_int)-1 &&
      attributes->mtime.seconds != (u_int)-1) {
    struct timeval t[2];
    t[0].tv_sec = attributes->atime.seconds;
    t[0].tv_usec = attributes->atime.useconds;
    t[1].tv_sec = attributes->mtime.seconds;
    t[1].tv_usec = attributes->mtime.useconds;
    if (utimes(filename, t) == -1) return -1;
  } else
#endif
  if (attributes->atime.seconds != (u_int)-1 ||
      attributes->mtime.seconds != (u_int)-1) {
    struct timespec t[2];
    t[0].tv_sec = attributes->atime.seconds;
    t[0].tv_nsec = attributes->atime.useconds * 1000;
    t[1].tv_sec = attributes->mtime.seconds;
    t[1].tv_nsec = attributes->mtime.useconds * 1000;
    if (attributes->atime.seconds == (u_int)-1) t[0].tv_nsec = UTIME_OMIT;
    else {
      while (t[0].tv_nsec >= 1000000000) {
        t[0].tv_nsec -= 1000000000;
        t[0].tv_sec++;
      }
    }
    if (attributes->mtime.seconds == (u_int)-1) t[1].tv_nsec = UTIME_OMIT;
    else {
      while (t[1].tv_nsec >= 1000000000) {
        t[1].tv_nsec -= 1000000000;
        t[1].tv_sec++;
      }
    }
printf("t[0] sec %ld nsec %ld\n", t[0].tv_sec, t[0].tv_nsec);
printf("t[1] sec %ld nsec %ld\n", t[1].tv_sec, t[1].tv_nsec);
    if (utimensat(AT_FDCWD, filename, t, AT_SYMLINK_NOFOLLOW) == -1) {
printf("utimensat error %s\n", strerror(errno));
return -1;
}
#if 0
printf("atime %d %d mtime %d %d\n",
  attributes->atime.seconds, attributes->atime.useconds,
  attributes->mtime.seconds, attributes->mtime.useconds);
    printf("%s:%d:%s TODO\n", __FILE__, __LINE__, __FUNCTION__);
    abort();
#endif
  }

  return 0;
}

void *
nfsproc_null_2_svc(void *argp, struct svc_req *rqstp)
{
  static char * result;
  printf("%s\n", __FUNCTION__);
  return (void *) &result;
}

attrstat get_stat(char *filename)
{
  struct stat s;
  attrstat result;

  if (lstat(filename, &s) == -1) goto errno_error;

  result.attrstat_u.attributes.type =
    S_ISREG(s.st_mode) ? NFREG :
    S_ISDIR(s.st_mode) ? NFDIR :
    S_ISBLK(s.st_mode) ? NFBLK :
    S_ISCHR(s.st_mode) ? NFCHR :
    S_ISLNK(s.st_mode) ? NFLNK :
    S_ISFIFO(s.st_mode) ? NFNON /* what to do? */ :
    S_ISSOCK(s.st_mode) ? NFNON :
      NFNON;
  result.attrstat_u.attributes.mode = s.st_mode;
  result.attrstat_u.attributes.nlink = s.st_nlink;
  result.attrstat_u.attributes.uid = s.st_uid;
  result.attrstat_u.attributes.gid = s.st_gid;
  result.attrstat_u.attributes.size = s.st_size;
  result.attrstat_u.attributes.blocksize = s.st_blksize;
  result.attrstat_u.attributes.rdev = s.st_rdev;
  result.attrstat_u.attributes.blocks = s.st_blocks;
  result.attrstat_u.attributes.fsid = s.st_dev;
  result.attrstat_u.attributes.fileid = s.st_ino;
  result.attrstat_u.attributes.atime.seconds = s.st_atim.tv_sec;
  result.attrstat_u.attributes.atime.useconds = s.st_atim.tv_nsec / 1000;
  result.attrstat_u.attributes.mtime.seconds = s.st_mtim.tv_sec;
  result.attrstat_u.attributes.mtime.useconds = s.st_mtim.tv_nsec / 1000;
  result.attrstat_u.attributes.ctime.seconds = s.st_ctim.tv_sec;
  result.attrstat_u.attributes.ctime.useconds = s.st_ctim.tv_nsec / 1000;

{
  fattr *attributes = &result.attrstat_u.attributes;
printf("get_stat %s\n", filename);
printf("mode 0x%x uid %d gid %d size %d\n",
  attributes->mode, attributes->uid, attributes->gid,
  attributes->size);
printf("atime %d %d mtime %d %d\n",
  attributes->atime.seconds, attributes->atime.useconds,
  attributes->mtime.seconds, attributes->mtime.useconds);
}

  result.status = 0;
  return result;

errno_error:
  result.status = errno2status(errno);
  return result;
}

attrstat *
nfsproc_getattr_2_svc(char *argp, struct svc_req *rqstp)
{
  static attrstat result;
  char *filename;

  printf("%s\n", __FUNCTION__);

  filename = fhandle2name(argp); if (filename == NULL) goto not_found;

  printf("filename %s\n", filename);

  result = get_stat(filename);
  return &result;

not_found:
  result.status = NFSERR_STALE;
  return &result;
}

attrstat *
nfsproc_setattr_2_svc(sattrargs *argp, struct svc_req *rqstp)
{
  char *filename;
  static attrstat  result;

  printf("%s\n", __FUNCTION__);

  filename = fhandle2name(argp->file); if (filename == NULL) goto not_found;

  printf("filename %s\n", filename);

  if (set_attributes(filename, &argp->attributes) == -1) goto errno_error;

  result = get_stat(filename);
  return &result;

errno_error:
  result.status = errno2status(errno);
printf("setattr error\n");
  return &result;

not_found:
  result.status = NFSERR_STALE;
  return &result;
}

void *
nfsproc_root_2_svc(void *argp, struct svc_req *rqstp)
{
  static char * result;
  printf("%s\n", __FUNCTION__);
  return (void *) &result;
}

diropres *
nfsproc_lookup_2_svc(diropargs *argp, struct svc_req *rqstp)
{
  char *dirname;
  char *filename;
  static diropres  result;
  unsigned long ul;
  attrstat attributes;

  printf("%s\n", __FUNCTION__);

  dirname = fhandle2name(argp->dir); if (dirname == NULL) goto not_found;

  filename = alloca(strlen(dirname) + 1 + strlen(argp->name) + 1);
  sprintf(filename, "%s/%s", dirname, argp->name);

printf("looking up %s\n", filename);

  /* this is brutal, huh? some hashtable or something needed here */
  /* bah, this is just a hack you know... */
  ul = new_file(filename);
  ul2fhandle(ul, result.diropres_u.diropok.file);

  attributes = get_stat(filename); if (attributes.status) goto stat_error;
  result.diropres_u.diropok.attributes = attributes.attrstat_u.attributes;

  result.status = 0;
  return &result;

stat_error:
  result.status = attributes.status;
  return &result;

not_found:
  result.status = NFSERR_STALE;
  return &result;
}

readlinkres *
nfsproc_readlink_2_svc(char *argp, struct svc_req *rqstp)
{
  /* is terminating 0 counted in NFS_MAXPATHLEN or not? */
  char *filename;
  static char link[NFS_MAXPATHLEN+1]; /* +1 or not? */
  ssize_t len;
  static readlinkres  result;

  printf("%s\n", __FUNCTION__);
  filename = fhandle2name(argp); if (filename == NULL) goto not_found;

  len = readlink(filename, link, NFS_MAXPATHLEN); /* -1 here? */
  if (len == -1) goto errno_error;
  link[len] = 0;

  result.readlinkres_u.data = link;
  return &result;

errno_error:
  result.status = errno2status(errno);
  return &result;

not_found:
  result.status = NFSERR_STALE;
  return &result;
}

readres *
nfsproc_read_2_svc(readargs *argp, struct svc_req *rqstp)
{
  static char buf[NFS_MAXDATA];
  char *filename;
  static readres  result;
  int fd;
  int len = NFS_MAXDATA;
  int ret;
  attrstat attributes;

  printf("%s\n", __FUNCTION__);

  filename = fhandle2name(argp->file); if (filename == NULL) goto not_found;
  fd = open(filename, O_RDONLY); if (fd == -1) goto errno_error;
  if (len > argp->count) len = argp->count;
  errno = 0;
  lseek(fd, argp->offset, SEEK_SET);
  if (errno) goto errno_error;
  ret = read(fd, buf, len);
  if (ret == -1) goto errno_error;
  close(fd); fd = -1;

  attributes = get_stat(filename); if (attributes.status) goto stat_error;
  result.readres_u.readok.attributes = attributes.attrstat_u.attributes;
  result.readres_u.readok.data.nfsdata_len = ret;
  result.readres_u.readok.data.nfsdata_val = buf;

  result.status = 0;
  return &result;

stat_error:
  result.status = attributes.status;
  return &result;

errno_error:
  result.status = errno2status(errno);
  if (fd != -1) close(fd);
  return &result;

not_found:
  result.status = NFSERR_STALE;
  return &result;
}

void *
nfsproc_writecache_2_svc(void *argp, struct svc_req *rqstp)
{
  static char * result;
  printf("%s\n", __FUNCTION__);
  return (void *) &result;
}

attrstat *
nfsproc_write_2_svc(writeargs *argp, struct svc_req *rqstp)
{
  char *filename;
  static attrstat  result;
  int fd;
  int ret;

  printf("%s\n", __FUNCTION__);

  filename = fhandle2name(argp->file); if (filename == NULL) goto not_found;
  fd = open(filename, O_WRONLY); if (fd == -1) goto errno_error;
  errno = 0;
  lseek(fd, argp->offset, SEEK_SET);
  if (errno) goto errno_error;
  ret = write(fd, argp->data.nfsdata_val, argp->data.nfsdata_len);
  if (ret == -1) goto errno_error;
  close(fd); fd = -1;

  result = get_stat(filename);
  return &result;

errno_error:
  result.status = errno2status(errno);
  if (fd != -1) close(fd);
  return &result;

not_found:
  result.status = NFSERR_STALE;
  return &result;
}

diropres *
nfsproc_create_2_svc(createargs *argp, struct svc_req *rqstp)
{
  char *dirname;
  char *filename;
  static diropres  result;
  unsigned long ul;
  attrstat attributes;

  printf("%s\n", __FUNCTION__);

  dirname = fhandle2name(argp->where.dir); if (dirname == NULL) goto not_found;

  filename = alloca(strlen(dirname) + 1 + strlen(argp->where.name) + 1);
  sprintf(filename, "%s/%s", dirname, argp->where.name);

printf("creating %s\n", filename);

  //if (creat(filename, 0755) == -1) goto errno_error;
  if (S_ISCHR(argp->attributes.mode) && argp->attributes.size == -1) {
    argp->attributes.mode &= ~(S_IFCHR);
    argp->attributes.mode |= S_IFIFO;
  }

  if (mknod(filename, argp->attributes.mode,
            argp->attributes.size) == -1) goto errno_error;

printf("mode 0x%x uid %d gid %d size %d\n",
  argp->attributes.mode, argp->attributes.uid, argp->attributes.gid,
  argp->attributes.size);
printf("atime %d %d mtime %d %d\n",
  argp->attributes.atime.seconds, argp->attributes.atime.useconds,
  argp->attributes.mtime.seconds, argp->attributes.mtime.useconds);

  if (set_attributes(filename, &argp->attributes) == -1) goto errno_error;

  /* this is brutal, huh? some hashtable or something needed here */
  /* bah, this is just a hack you know... */
  ul = new_file(filename);
  ul2fhandle(ul, result.diropres_u.diropok.file);

  attributes = get_stat(filename); if (attributes.status) goto stat_error;
  result.diropres_u.diropok.attributes = attributes.attrstat_u.attributes;

  result.status = 0;
  return &result;

stat_error:
  result.status = attributes.status;
  unlink(filename);
  return &result;

errno_error:
  result.status = errno2status(errno);
  unlink(filename);
printf("create error errno %d %s\n", errno, strerror(errno));
  return &result;

not_found:
  result.status = NFSERR_STALE;
  return &result;
}

nfs_stat *
nfsproc_remove_2_svc(diropargs *argp, struct svc_req *rqstp)
{
  char *dirname;
  char *filename;
  static nfs_stat  result;

  printf("%s\n", __FUNCTION__);

  dirname = fhandle2name(argp->dir); if (dirname == NULL) goto not_found;
  filename = alloca(strlen(dirname) + 1 + strlen(argp->name) + 1);
  sprintf(filename, "%s/%s", dirname, argp->name);

  if (unlink(filename) == -1) goto errno_error;

  result = 0;
  return &result;

errno_error:
  result = errno2status(errno);
  return &result;

not_found:
  result = NFSERR_STALE;
  return &result;
}

nfs_stat *
nfsproc_rename_2_svc(renameargs *argp, struct svc_req *rqstp)
{
  char *from_dirname;
  char *from_filename;
  char *to_dirname;
  char *to_filename;
  static nfs_stat  result;

  printf("%s\n", __FUNCTION__);

  from_dirname = fhandle2name(argp->from.dir);
  if (from_dirname == NULL) goto not_found;
  to_dirname = fhandle2name(argp->to.dir);
  if (to_dirname == NULL) goto not_found;

  from_filename = alloca(strlen(from_dirname) +1+ strlen(argp->from.name)+1);
  sprintf(from_filename, "%s/%s", from_dirname, argp->from.name);
  to_filename = alloca(strlen(to_dirname) +1+ strlen(argp->to.name)+1);
  sprintf(to_filename, "%s/%s", to_dirname, argp->to.name);

  if (rename(from_filename, to_filename) == -1) goto errno_error;

  result = 0;
  return &result;

errno_error:
  result = errno2status(errno);
  return &result;

not_found:
  result = NFSERR_STALE;
  return &result;
}

nfs_stat *
nfsproc_link_2_svc(linkargs *argp, struct svc_req *rqstp)
{
  char *from;
  char *filename;
  char *dirname;
  static nfs_stat  result;

  printf("%s\n", __FUNCTION__);

  from = fhandle2name(argp->from); if (from == NULL) goto not_found;

  dirname = fhandle2name(argp->to.dir); if (dirname == NULL) goto not_found;

  filename = alloca(strlen(dirname) + 1 + strlen(argp->to.name) + 1);
  sprintf(filename, "%s/%s", dirname, argp->to.name);

  if (link(from, filename) == -1) goto errno_error;

  result = 0;
  return &result;

errno_error:
  result = errno2status(errno);
  return &result;

not_found:
  result = NFSERR_STALE;
  return &result;
}

nfs_stat *
nfsproc_symlink_2_svc(symlinkargs *argp, struct svc_req *rqstp)
{
  char *dirname;
  char *filename;
  static nfs_stat  result;

  printf("%s\n", __FUNCTION__);

  dirname = fhandle2name(argp->from.dir); if (dirname == NULL) goto not_found;

  filename = alloca(strlen(dirname) + 1 + strlen(argp->from.name) + 1);
  sprintf(filename, "%s/%s", dirname, argp->from.name);

printf("filename %s target %s\n", filename, argp->to);

  if (symlink(argp->to, filename) == -1) goto errno_error;

  if (set_attributes(filename, &argp->attributes) == -1) goto errno_error;

  result = 0;
  return &result;

errno_error:
  result = errno2status(errno);
  unlink(filename);
  return &result;

not_found:
  result = NFSERR_STALE;
  return &result;
}

diropres *
nfsproc_mkdir_2_svc(createargs *argp, struct svc_req *rqstp)
{
  char *dirname;
  char *filename;
  static diropres  result;
  unsigned long ul;
  attrstat attributes;

  printf("%s\n", __FUNCTION__);

  dirname = fhandle2name(argp->where.dir); if (dirname == NULL) goto not_found;

  filename = alloca(strlen(dirname) + 1 + strlen(argp->where.name) + 1);
  sprintf(filename, "%s/%s", dirname, argp->where.name);

  if (mkdir(filename, 0755) == -1) goto errno_error;

printf("mode 0x%x uid %d gid %d size %d\n",
  argp->attributes.mode, argp->attributes.uid, argp->attributes.gid,
  argp->attributes.size);
printf("atime %d %d mtime %d %d\n",
  argp->attributes.atime.seconds, argp->attributes.atime.useconds,
  argp->attributes.mtime.seconds, argp->attributes.mtime.useconds);

  if (set_attributes(filename, &argp->attributes) == -1) goto errno_error;

  /* this is brutal, huh? some hashtable or something needed here */
  /* bah, this is just a hack you know... */
  ul = new_file(filename);
  ul2fhandle(ul, result.diropres_u.diropok.file);

  attributes = get_stat(filename); if (attributes.status) goto stat_error;
  result.diropres_u.diropok.attributes = attributes.attrstat_u.attributes;

  result.status = 0;
  return &result;

stat_error:
  result.status = attributes.status;
  rmdir(filename);
  return &result;

errno_error:
  result.status = errno2status(errno);
  rmdir(filename);
  return &result;

not_found:
  result.status = NFSERR_STALE;
  return &result;
}

nfs_stat *
nfsproc_rmdir_2_svc(diropargs *argp, struct svc_req *rqstp)
{
  char *dirname;
  char *filename;
  static nfs_stat  result;

  printf("%s\n", __FUNCTION__);

  dirname = fhandle2name(argp->dir); if (dirname == NULL) goto not_found;
  filename = alloca(strlen(dirname) + 1 + strlen(argp->name) + 1);
  sprintf(filename, "%s/%s", dirname, argp->name);

  if (rmdir(filename) == -1) goto errno_error;

  result = 0;
  return &result;

errno_error:
  result = errno2status(errno);
  return &result;

not_found:
  result = NFSERR_STALE;
  return &result;
}

readdirres *
nfsproc_readdir_2_svc(readdirargs *argp, struct svc_req *rqstp)
{
  static entry e[NFS_MAXDATA];
  static int e_inited = 0;
  int i;
  char *filename;
  static readdirres  result;
  int start;
  int curlen;
  DIR *d;
  int eof;
  cookie_cache *nfscookies;

  if (e_inited == 0) { e_inited = 1; bzero(e, sizeof(e)); }

  printf("%s\n", __FUNCTION__);

  filename = fhandle2name(argp->dir); if (filename == NULL) goto not_found;
  nfscookies = fhandle2cookies(argp->dir); /* no test needed */
  start = *(int *)argp->cookie; /* not portable */

printf("start = %d\n", start);
fflush(stdout);

  d = opendir(filename); if (d == NULL) goto errno_error;
  if (start) seekdir(d, get_cookie(nfscookies, start));
  else clear_cookie_cache(nfscookies);

  curlen = 4 /* status */ + 4 /* null entry */ + 4 /* eof indicator */;

  eof = 0;
  for (i = 0; i < NFS_MAXDATA; i++) {
    struct dirent *r;
    errno = 0;
    r = readdir(d);
    int rsize;
    if (r == NULL && errno) goto errno_error;
    if (r == NULL) { eof = 1; break; }
    rsize = 4 /* fileid */ + 4 /* name length */
           + ((strlen(r->d_name) + 3) & ~ 3) /* length of name rounded up */
           + 4 /* cookie */ + 4 /* nextentry */;
    if (curlen + rsize > argp->count) goto full;
    if (i) { e[i-1].nextentry = &e[i]; }
    if (e[i].name) free(e[i].name);
    e[i].fileid = r->d_ino;
    e[i].name = strdup(r->d_name); if (e[i].name == NULL) abort();
    *(int *)e[i].cookie = add_cookie(nfscookies, r->d_off); /* not portable */
printf("  %s d_off %ld\n", r->d_name, r->d_off);
    e[i].nextentry = NULL;
    curlen += rsize;
  }

  /* maybe it's eof
   * (I doubt this is ever reached, the for above does
   * 'goto full' when done with eof still==0, NFS_MAXDATA is too big) */
  if (!eof) {
    struct dirent *r;
    r = readdir(d);
    if (r == NULL && errno) goto errno_error;
    if (r == NULL) eof = 1;
  }

full:

printf("readdir done (%d entries filled)\n", i);
fflush(stdout);

  if (closedir(d) == -1) { d = NULL; goto errno_error; }

  result.readdirres_u.readdirok.entries = e;
  result.readdirres_u.readdirok.eof = eof;

  result.status = 0;
  return &result;

errno_error:
  result.status = errno2status(errno);
  if (d != NULL) closedir(d);
  return &result;

not_found:
  result.status = NFSERR_STALE;
  return &result;
}

statfsres *
nfsproc_statfs_2_svc(char *argp, struct svc_req *rqstp)
{
  struct statvfs s;
  static statfsres  result;
  char *filename;

  printf("%s\n", __FUNCTION__);

  filename = fhandle2name(argp); if (filename == NULL) goto not_found;
  if (statvfs(filename, &s) == -1) goto errno_error;

  result.statfsres_u.info.tsize = s.f_frsize;  /* what to put? */
  result.statfsres_u.info.bsize = s.f_frsize;
  result.statfsres_u.info.blocks = s.f_blocks;
  result.statfsres_u.info.bfree = s.f_bfree;
  result.statfsres_u.info.bavail = s.f_bavail;

  result.status = 0;
  return &result;

errno_error:
  result.status = errno2status(errno);
  return &result;

not_found:
  result.status = NFSERR_STALE;
printf("not found\n");
  return &result;
}
